from collections import defaultdict
from typing import List

from django.db.models.sql import compiler
from django.db.models.sql.compiler import SQLCompiler
from django.db.models.sql.constants import INNER
from django.db.models.sql.datastructures import Join

from .exception import JoinError


class JoinField:

    def __init__(self, columns: List[str]):
        self._joining_columns = columns

    def get_joining_columns(self) -> List[str]:
        return self._joining_columns

    def get_extra_restriction(self, alias: str, remote_alias: str, *args, **kwargs) -> None:
        pass


class SQLUnrelatedJoinCompiler(SQLCompiler):
    join_cls = Join

    def setup_unrelated_joins(self, join_type: str = INNER, nullable: bool = False, **join_fields) -> None:
        self.setup_query()

        if not self.query.model:
            raise JoinError('Unable to setup joins with no parent model.')

        tables = defaultdict(list)

        for from_field, to_field in join_fields.items():
            alias = to_field.field.model._meta.db_table
            tables[alias].append((from_field, to_field.field.column))

        for alias, fields in tables.items():
            join = self.join_cls(
                table_name=alias,
                parent_alias=self.query.model._meta.db_table,
                table_alias=alias,
                join_type=join_type,
                join_field=JoinField(fields),  # type: ignore[arg-type]
                nullable=nullable,
            )
            self.query.alias_map[alias] = join
            self.query.alias_refcount[alias] = 1


setattr(compiler, 'SQLUnrelatedJoinCompiler', SQLUnrelatedJoinCompiler)



from typing import Dict, Optional, Type

from django.db.models import Field, Model, QuerySet
from django.db.models.sql import Query
# just to keep attention to this constants
from django.db.models.sql.constants import INNER, LOUTER

from .compiler import SQLUnrelatedJoinCompiler
from .exception import JoinError


class JoinType:
    LEFT = LOUTER
    RIGHT = 'RIGHT OUTER JOIN'
    INNER = INNER
    FULL = 'FULL OUTER JOIN'


class JoinQuery(Query):
    compiler = 'SQLUnrelatedJoinCompiler'
    alias_prefix = 'J'

    def __init__(self, model: Optional[Type[Model]], alias_cols: bool = True):
        super().__init__(model=model, alias_cols=alias_cols)
        self._join_type: str = ''
        self._join_nullable: bool = False
        self._join_fields: Dict[str, Field] = {}

    def add_join_values(self, join_type: str = JoinType.INNER, nullable: bool = False, **join_fields) -> None:
        self._join_type = join_type
        self._join_nullable = nullable
        self._join_fields = join_fields

    def get_compiler(self, *args, **kwargs) -> SQLUnrelatedJoinCompiler:
        comp: SQLUnrelatedJoinCompiler = super().get_compiler(*args, **kwargs)  # type: ignore[assignment]
        if self._join_fields:
            comp.setup_unrelated_joins(join_type=self._join_type, nullable=self._join_nullable, **self._join_fields)
        return comp


class UnrelatedJoinQuerySet(QuerySet):

    def join(self, join_type=JoinType.INNER, nullable=False, **join_fields) -> 'UnrelatedJoinQuerySet':
        # can't be joined on itself
        # for FK use *_id field: id=Person.department_id
        for i in join_fields.values():
            if i.field.model == self.model:
                raise JoinError('Unable to join on the same model.')

        self.query: JoinQuery = self.query.chain(JoinQuery)  # type: ignore[assignment]
        self.query.add_join_values(join_type=join_type, nullable=nullable, **join_fields)
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_join_unrelated']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'django-join-unrelated',
    'version': '0.2.0',
    'description': 'Use SQL Join with Django ORM models having no relations.',
    'long_description': '# django-join-unrelated\n![Python version](https://img.shields.io/badge/Python-3.8%2B-blue)\n[![Tests](https://github.com/KazakovDenis/django-join-unrelated/actions/workflows/cicd.yml/badge.svg)](https://github.com/KazakovDenis/django-join-unrelated/actions/workflows/cicd.yml)\n\n### Use SQL Join with Django ORM models having no relations.\n\nIf you have models in your project that share the same data but have no relations, \nyou might face a situation when you need to join them. But Django does not provide \nthis functionality yet.\n  \nWith `django-join-unrelated` you can do the following \n(models defined [here](https://github.com/KazakovDenis/django-join-unrelated/blob/main/app/core/models.py)):\n```python\nfrom app.core.models import Jedi, Person\n\nPerson.objects.create(first_name=\'Padme\', last_name=\'Amidala\', birth_place=\'Naboo\')\nPerson.objects.create(first_name=\'Obi-Wan\', last_name=\'Kenobi\', birth_place=\'Stewjon\')\nJedi.objects.create(first_name=\'Obi-Wan\', last_name=\'Kenobi\', force=100)\nJedi.objects.create(first_name=\'Mace\', last_name=\'Windu\', force=80)\n\nJedi.objects.join(first_name=Person.first_name).values_list(\'first_name\')\n# <UnrelatedJoinQuerySet [(\'Obi-Wan\',)]>\n\nprint(Jedi.objects.join(first_name=Person.first_name).values_list(\'first_name\').query)\n# SELECT "core_jedi"."first_name" FROM "core_jedi" INNER JOIN core_person ON ("core_jedi"."first_name" = core_person."first_name")\n\n# Set join type explicitly\nfrom django_join_unrelated import JoinType\n\nJedi.objects.join(first_name=Person.first_name, join_type=JoinType.FULL)\n\n# Join on multiple fields\nJedi.objects.join(first_name=Person.first_name, last_name=Person.last_name)\n\n# Join multiple models\nJedi.objects.join(first_name=Person.first_name, id=User.id)\n```\n\n`django-join-unrelated` tries to keep all QuerySet power where it is possible:\n```python\nJedi.objects.join(first_name=Person.first_name).filter(last_name=\'Kenobi\')\n# <UnrelatedJoinQuerySet [<Jedi: Obi-Wan Kenobi>]>\n\nJedi.objects.join(first_name=Person.first_name).annotate(name=F(\'first_name\'))[0].name\n# \'Obi-Wan\'\n\nJedi.objects.join(first_name=Person.first_name).aggregate(Sum(\'force\'))\n# {\'force__sum\': 100}\n```\n\n## Installation\nYou can install `django-join-unrelated` using pip:\n\n```\npip install django-join-unrelated\n```\n\n## Usage\nJust import `UnrelatedJoinManager` and set it as an attribute to your models:\n\n```python\nfrom django.db import models\nfrom django_join_unrelated import UnrelatedJoinManager\n\nclass Jedi(models.Model):\n    first_name = models.CharField(\'First name\', max_length=128)\n    last_name = models.CharField(\'Last name\', max_length=128)\n    force = models.IntegerField(\'Force\')\n\n    objects = UnrelatedJoinManager()\n\n    class Meta:\n        verbose_name = \'Jedi\'\n```\n  \n**Already implemented:**\n- filtering via joins\n\n**Not implemented yet but planned:**\n- unrelated objects selection\n- return of joined values\n',
    'author': 'Denis Kazakov',
    'author_email': 'denis@kazakov.ru.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/django-join-unrelated',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
